import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
print(f"PYTORCH VERSION = {torch.__version__}")

import torchvision
from torchvision import transforms
# import torchvision.datasets as datasets
from torchvision import datasets
from torchvision.transforms import ToTensor
from torchvision.transforms.functional import InterpolationMode

# import ffutils

import anyact as aa

import h5py

from toy_problem_mnist_analysis import analysis_classification
from toy_problem_mnist import fmnist,mnist,mnist_perf

import matplotlib.pyplot as plt
# plt.ion()

π = np.pi
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu') # use GPU 0 only
torch.cuda.empty_cache()

# def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

#################################

def show(im,normalize=False,normalize_range=None,path=None):
	from torchvision.utils import make_grid
	# if len(im.shape)==4: im = [i for i in im]
	# if not isinstance(im, list): im = [im]
	# im = make_grid(im)
	im = make_grid(im, nrow = int(np.ceil(np.sqrt(len(im)))), normalize=normalize, value_range=normalize_range)
	img = torchvision.transforms.functional.to_pil_image(im.detach())
	if path==None: plt.imshow(np.asarray(img))
	else:          img.save(path)

import SOLO_shared

def pad_image(image, target_resolution):
	original_height, original_width = image.size()[-2:]
	target_height, target_width = target_resolution
	# Calculate the required amount of padding
	pad_height = max(target_height - original_height, 0)
	pad_width = max(target_width - original_width, 0)
	# Compute the padding amounts for the top, bottom, left, and right sides
	pad_T = pad_height // 2
	pad_B = pad_height - pad_T
	pad_L = pad_width // 2
	pad_R = pad_width - pad_L
	return F.pad(image, (pad_L, pad_R, pad_T, pad_B))

class InputAndParam(nn.Module):
	def to(self,device):
		print(f"to device {device}")
		nn.Module.to(self,device)
		self.mask = self.mask.to(device)
		return self
	def __init__(self,stupid_mode=False):
		super().__init__()
		resx=resy=28
		self.stupid_mode=stupid_mode
		self.mask = (1 - F.pad(SOLO_shared.SOLO_conv(torch.ones(1,resy,resx),stride=2), [1,1,1,1], "constant", 0)).bool()
		# self.θ = nn.parameter.Parameter(torch.rand((1,2*resy+1,2*resx+1)))
		if not self.stupid_mode: self.θ = nn.parameter.Parameter(torch.zeros((1,2*resy+1,2*resx+1)))
	def forward(self,x):
		if not self.stupid_mode:
			xx = SOLO_shared.SOLO_conv(x,stride=2)              # blow it
			xx = F.pad(xx, [1,1,1,1], "constant", 0) # pad it
			θθ = .5+.5*torch.sin(self.θ) # θ must be in [0,1]...
			out = θθ*self.mask + xx # combine with weights
			return pad_image(out, [64,64])
		else:
			xx = SOLO_shared.SOLO_conv(x,stride=2)              # blow it
			xx = F.pad(xx, [1,1,1,1], "constant", 0) # pad it
			out = xx
			return pad_image(out, [64,64])

# fmnist_L_x,fmnist_L_y = fmnist(True)
# fmnist_T_x,fmnist_T_y = fmnist(False)
# x = fmnist_T_x[:12]
# im = InputAndParam()(x)
# import matplotlib
# matplotlib.use('qtagg')
# show(im)
# plt.show()
# show(im,path="./out/SOLO_InputEncode.png")

# test = nn.Sequential(
# 	InputAndParam().to(device),
# 	SOLO_mini.SOLO_derivable(opti_mode,bwd_lr,loss_mode,identifier,bwd_sampling=dim_sample,stupid_mode=stupid_mode).to(device),
# 	# nn.Flatten(),
# 	# nn.Linear(64*64,10),
# 	# nn.Softmax(1),
# ).to(device)

# im = test(x)
# show(im)
# plt.show()


import SOLO_mini

stupid_input=True
stupid_mode = True
dim_sample_pow=-18 #		for dim_sample_pow in [-18,-20,-22,-24]:
# dim_sample_pow=-22 #		for dim_sample_pow in [-18,-20,-22,-24]:

for stupid_input in [True,False]:
	for stupid_mode in [True,False]:
		for dim_sample_pow in [-24,-20,-18]:
			dim_sample=2**dim_sample_pow
			loss_mode="mse"
			opti_mode="sgd"
			bwd_lr=1e-3
			identifier = f"dim_sample=2^{dim_sample_pow}-loss={loss_mode}-opti={opti_mode}-bwd_lr={bwd_lr}{'-stupid' if stupid_mode else ''}{'-stupidInput' if stupid_input else ''}"
			add_job(lambda:
				analysis_classification(
					nn.Sequential(
						InputAndParam(stupid_input).to(device),
						SOLO_mini.SOLO_derivable(opti_mode,bwd_lr,loss_mode,identifier,bwd_sampling=dim_sample,stupid_mode=stupid_mode).to(device),
						nn.Flatten(),
						nn.Linear(64*64,10),
						nn.Softmax(1),
					).to(device),
					identifier = identifier,
					loss_fun=nn.MSELoss(),
					lr0=-2, # external learning rate
					lr1=-4, # external learning rate
					epochs=128,
					batch_size=64,
					# batch_size=256,
					mini_data=True
				)
			)











































# ###################
# ###### DEBUG ######
# ###################

# # class InputAndParam(nn.Module):
# # 	def __init__(self,θ_n):
# # 		super().__init__()
# # 		self.θ = nn.parameter.Parameter(torch.rand(θ_n))
# # 		self.flatten = nn.Flatten()
# # 	def forward(self,x):
# # 		xx = self.flatten(x)
# # 		θθ = .5+.5*torch.sin(self.θ) # θ must be in [0,1]...
# # 		out = torch.hstack((xx,θθ.repeat(xx.shape[0],1))) # combine with weights
# # 		return out

# # x = fmnist_T_x[:12]
# # θ_n = 28*28*3
# # im = InputAndParam(θ_n)(x)
# # # show(im,path="./out/SOLO_InputEncode.png")
# # show(im)


# # with ffutils.NaNDetect():
# identifier = f"DEBUG"
# θ_n = 28*28*3
# analysis_classification(
# 	nn.Sequential(
# 		InputAndParam().to(device), nn.Flatten(), nn.Linear(64*64,10),
# 		# InputAndParam(θ_n), nn.Flatten(), nn.Linear(28*28+θ_n,10),
# 		# nn.Flatten(), nn.Linear(28*28,10),
# 		nn.Softmax(1),
# 	),
# 	identifier = identifier,
# 	loss_fun = nn.MSELoss(),
# 	lr0=-2, # external learning rate
# 	lr1=-4, # external learning rate
# 	epochs=128,
# 	batch_size=64,
# 	# batch_size=256,
# )



# # fmnist_L_x,fmnist_L_y = fmnist(True)
# # fmnist_T_x,fmnist_T_y = fmnist(False)

# # x = fmnist_T_x[:10]
# # im = InputAndParam()(x)
# # show(im)










# # InputAndParam()(torch.randn(10,1,28,28)).shape

# import SOLO_001_small

# ########################################################
# ###### EXTERNAL PROBLEM : blackbox(x,θ) ≈ toy_A⋅x ######
# ########################################################

# ### TODO : don't end until we haven't reached model.bwd_epoch_rel > epochs_rel

# # epochs = epochs_rel # we will end before as bwd_epoch_rel is the ending criterion for now

# # blackbox = Blackbox(opti_mode,bwd_lr,loss_mode,identifier,stupid_mode=True).to(device)
# # out_shape = blackbox(torch.randn(10,1,28,28).to(device)).shape
# # out_dim = out_shape[-1]*out_shape[-2]
# # model=nn.Sequential(
# # 	blackbox,
# # 	nn.Flatten(),
# # 	nn.Linear(out_dim,10),
# # ).to(device)

# ### DEBUG
# # model = nn.Sequential(
# # 	nn.Flatten(),
# # 	aa.make_nn_fully_connected_real(28*28,10,10),
# # ).to(device)


# for loss_mode in ["mse","mse+cos","mse,cos"]:
# 	for opti_mode,bwd_lr in [
# 		["adam",1e-5],
# 		["adam",1e-4],
# 		["adam",1e-3],
# 		["adam",1e-2],
# 	]:
# 		# for dim_sample in 2.**np.array([0,-2,-4,-6,-8]):
# 		for dim_sample in 2.**np.array([-18,-20,-22,-24]):
# 				# for stupid_mode in [False,True]:
# 				for stupid_mode in [True]:
# 					# with ffutils.NaNDetect():
# 						identifier = f"dim_sample={dim_sample}-loss={loss_mode}-opti={opti_mode}-bwd_lr={bwd_lr}-stupid={stupid_mode}"
# 						add_job(lambda:
# 							analysis_classification(
# 								nn.Sequential(
# 									InputAndParam().to(device),
# 									SOLO_001_small.SOLO_derivable(opti_mode,bwd_lr,loss_mode,identifier,bwd_sampling=dim_sample,stupid_mode=stupid_mode).to(device),
# 									nn.Flatten(),
# 									nn.Linear(64*64,10),
# 									nn.Softmax(1),
# 								).to(device),
# 								identifier = identifier,
# 								loss_fun=nn.MSELoss(),
# 								lr0=-2, # external learning rate
# 								lr1=-4, # external learning rate
# 								epochs=128,
# 								batch_size=64,
# 								# batch_size=256,
# 							)
# 						)

# # def calc_cnn(im_size,ker_size=3,maxpool_size=2): return (im_size-ker_size+1)//maxpool_size

# # for k in [1]:
# # 	identifier = f"DEBUG"
# # 	mnist_analysis.analysis_classification(
# # 		# nn.Sequential(
# # 		# 	InputAndParam().to(device),
# # 		# 	SOLO_001_small.SOLO_derivable(opti_mode,bwd_lr,loss_mode,identifier,bwd_sampling=dim_sample,stupid_mode=stupid_mode).to(device),
# # 		# 	nn.Flatten(),
# # 		# 	nn.Linear(64*64,10),
# # 		# 	nn.Softmax(1),
# # 		# ).to(device),
# # 		nn.Sequential(
# # 			nn.Conv2d(1  ,128,3),nn.GELU(),nn.MaxPool2d(2),nn.BatchNorm2d(128),
# # 			nn.Conv2d(128,128,3),nn.GELU(),nn.MaxPool2d(2),nn.BatchNorm2d(128),
# # 			nn.Conv2d(128,128,3),nn.GELU(),nn.MaxPool2d(2),nn.BatchNorm2d(128),
# # 			nn.Flatten(),
# # 			# nn.Linear(64*(calc_cnn(28)**2),10),nn.GELU(),
# # 			# nn.Linear(64*(calc_cnn(calc_cnn(28))**2),10),nn.GELU(),
# # 			nn.Linear(128*(calc_cnn(calc_cnn(calc_cnn(28)))**2),64),nn.GELU(),
# # 			nn.Linear(64,10),nn.GELU(),
# # 			nn.Softmax(1),
# # 			# nn.Flatten(),
# # 			# nn.Linear(28*28,64),nn.GELU(),
# # 			# nn.Linear(64,10),
# # 			# nn.Softmax(1),
# # 		),
# # 		identifier = identifier,
# # 		loss_fun = nn.MSELoss(),
# # 		lr0=-2, # external learning rate
# # 		lr1=-4, # external learning rate
# # 		epochs=128,
# # 		# batch_size=64,
# # 		batch_size=256,
# # 	)

